import logging
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)
import os
import json
import dill
import pandas as pd

#### tools
def save_thunder(obj, path, filename='thunder.p'):
    dill.dump(obj, open(os.path.join(path, filename), 'wb'))

def load_thunder(path):
    obj = dill.load(open(path, 'rb'))
    return obj

def save_plot(plot, path='.', figname='figure.png'):
    plot.savefig(os.path.join(path, figname), transparent=True, format='svg')

def save_fit_report(obj, path, filename="report.json"):
    json.dump(obj, open(os.path.join(path, filename), 'w'))



def find_closest_indices(list1, list2):
    list_of_matching_indices = [min(range(len(list1)), key=lambda i: abs(list1[i] - cent))
                                for cent in list2]
    return list_of_matching_indices
#### tools

### user inputs and loading etc
def tightness_setter(tightness):
    tight_dict = {}
    tight_dict['centre_bounds'] = 1
    tight_dict['width_bounds'] = (5, 2)
    tight_dict['amps_bounds'] = (2, 2)

    if tightness == 'low':
        tight_dict['centre_bounds'] = 10
        tight_dict['width_bounds'] = (20, 3)
        tight_dict['amps_bounds'] = (5, 3)
    elif tightness == "med":
        pass
    elif tightness == 'high':
        tight_dict['centre_bounds'] = 0.5
        tight_dict['width_bounds'] = (2, 1)
        tight_dict['amps_bounds'] = (1.2, 1.2)

    else:
        logging.warning(
            'The tightness defined was incorrect format, use low, med or high. Using default med settings')

    return tight_dict

def load_data(datapath, x_ind, y_ind, e_ind=None):
    """
    load in data as a pandas df - save by modifying self.data, use object params to load
    :return: None
    """
    if '.h5' in datapath: # if the data is already stored as a pandas df
        store = pd.HDFStore(datapath)
        keys = store.keys()
        if len(keys) > 1:
            LOGGER.warning("Too many keys in the hdfstore, will assume all should be concated")
            LOGGER.warning("not sure this concat works yet")
            data = store.concat([store[key] for key in keys]) # not sure this will work! concat all keys dfs together
        else:
            data = store[keys[0]] # if only one key then we use it as the datafile
    else: # its a txt or csv file
        data = pd.read_csv(datapath, header=None, sep='\t') # load in, works for .txt and .csv
        # this needs to be made more flexible/user defined

    if e_ind: # if we have specified this column then we use it, otherwise just x and y
        assert (len(data.columns) >= 2), "You have specified an e_ind but there are less than 3 columns in the data"
        e_data = data[e_ind].values
    else:
        e_data = None

    data.dropna() # drop any rows with NaN etc in them

    x_data = data[x_ind].values
    y_data = data[y_ind].values

    return x_data, y_data, e_data

def parse_param_file(filepath='./params.txt'):
    """
    parse a params file which we assume is a dictionary
    :param filepath: str: path to params file
    :return: dictionary of paramters
    """
    # maybe use json loads if you end up writing parameter files non-manually

    with open(filepath, 'r') as f:
        arguments = json.load(f)
        f.close()

    # TODO: add some checks to user passed data
    return arguments

def parse_args(arg):
    """
    convert argparse arguments into a dictionary for consistency later
    :param arg: argparse parsed args
    :return: dictionary of parameters
    """
    arguments = {}
    arguments['x_ind'] = arg.x_ind
    arguments['y_ind'] = arg.y_ind
    arguments['e_ind'] = arg.e_ind
    arguments['datapath'] = arg.datapath
    arguments['no_peaks'] = arg.no_peaks
    arguments['background'] = arg.background
    arguments['scarf_params'] = arg.scarf_params
    arguments['peak_types'] = arg.peak_types
    arguments['peak_centres'] = arg.peak_centres
    arguments['peak_widths'] = arg.peak_widths
    arguments['peak_amps'] = arg.peak_amps
    arguments['tightness'] = arg.tightness
    arguments['bounds'] = arg.bounds

    # TODO: add some checks to user passed data
    return arguments

def make_dir(dirname, i=1):
    """
    function to make a directory, recursively adding _new if that name already exists
    :param dirname: str: name of directory to create
    :param i: the run number we are on
    :return: str: the directory name which was available, and all subsequent data should be saved in
    """
    try:
        os.mkdir(f'{dirname}')
    except FileExistsError as e:
        dirname = make_dir(f'{dirname}_new', i + 1)
        if i == 1:
            print(e, f'. So I named the file: {dirname}')
        return dirname
    return dirname
####




import logging
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)
import difflib
import re
import numpy as np

from typing import Dict, Union
import copy

import matplotlib
matplotlib.use('TkAgg')
import matplotlib.pyplot as plt

import lmfit

from . import utilities as utili
from . import plotting


# TODO: need to fail if peak fitting doesn't work!

class Thunder():
    """
    thunder object with all the methods we love inside it. Name generated using WuTang Clan name generator.
    """
    def __init__(self, input):
        self.input: Union[Thunder, Dict] = input

        self.x_ind: int = 0
        self.y_ind: int = 1
        self.e_ind: Union[int, None] = None

        self.x_data: Union[None, np.ndarray] = None
        self.y_data: Union[None, np.ndarray] = None
        self.e_data: Union[None, np.ndarray] = None

        self.y_data_bg_rm = None
        self.y_data_norm = None

        self.datapath: str = './data.txt'

        self.no_peaks: int = 0
        self.background: str = "SCARF"
        self.scarf_params: Union[None, Dict] = None
        self.peak_types: Union[None, list] = []
        self.peak_centres: Union[None, list] = []
        self.peak_widths: Union[None, list] = []
        self.peak_amps: Union[None, list] = []
        self.tightness: str = "med"
        self.bounds: Union[None, Dict] = None

        self.peaks: lmfit.model.ModelResult
        self.plot: plt = None
        self.fit_report: {} = {}

        self.free_params: int = 0
        self.p_value: int = 0
        self.chisq: int = 0

        if isinstance(input, Thunder):  # if only pass one but its already a thunder object then just use that
            self.overwrite_thunder(input)  # add all the details in depending on args
        elif isinstance(input, dict):
            self.create_thunder(input)  # add all the details in depending on args
        else:
            raise TypeError('Cannot convert input to Thunder object')

        if not self.x_data and not self.y_data:
            self.x_data, self.y_data, self.e_data = utili.load_data(self.datapath, self.x_ind, self.y_ind) # load the data

        self.tightness = utili.tightness_setter(self.tightness)

    def overwrite_thunder(self, inp):
        thun = inp

        if thun.x_data and thun.y_data:
            self.x_data = thun.x_data
            self.y_data = thun.y_data
        else:
            self.x_ind = thun.x_ind
            self.y_ind = thun.y_ind
            self.e_ind = thun.e_ind
            self.datapath = thun.datapath

        if thun.y_data_bg_rm:
            self.y_data_bg_rm = thun.y_data_bg_rm
        if thun.y_data_norm:
            self.y_data_norm =thun.y_data_norm

        self.no_peaks = thun.no_peaks
        self.background = thun.background
        self.scarf_params = thun.scarf_params
        self.peak_types = thun.peak_types
        self.peak_centres = thun.peak_centres
        self.peak_widths = thun.peak_widths
        self.peak_amps = thun.peak_amps
        self.tightness = thun.tightness
        self.bounds = thun.bounds

    def create_thunder(self, inp: Dict):
        """
        Used to create a thunder object given different input types
        :param args: a,b,c depending on type of input and
        :return: None, we modify the object unless a spec1d object is passed, in which case we return that
        """
        try: # only continue if its e_ind missing
            self.e_ind = inp['e_ind']
        except KeyError as e:
            LOGGER.info(f"KeyError: Missing field in the data dictionary: {e}")

        try: # If its the others you need to fail here
            self.datapath = inp['datapath']
            self.x_ind = inp['x_ind']
            self.y_ind = inp['y_ind']
        except KeyError as e:
            raise KeyError(f"Missing vital information to load object: {e}")

        self.no_peaks = inp.get('no_peaks', self.no_peaks)
        self.background = inp.get('background', self.background)
        # do some check on background here to set it to an np array
        self.scarf_params = inp.get('scarf_params', self.scarf_params)
        self.peak_types = inp.get('peak_types', self.peak_types)
        self.peak_centres = inp.get('peak_centres', self.peak_centres)
        self.peak_widths = inp.get('peak_widths', self.peak_widths)
        self.peak_amps = inp.get('peak_amps', self.peak_amps)
        self.tightness = inp.get('tightness', self.tightness)
        self.bounds = inp.get('bounds', self.bounds)

    ## plot_all and fit_report need imporovements e.g. to check which attributes exists in the object
    def plot_all(self):
        ax = plotting.plot_fits(self.x_data, self.peaks.eval_components()) # plot each component of the model
        ax = plotting.plot_background(self.x_data, self.background, ax) #plot the background supplied by user
        ax = plotting.plot_fit_sum(self.x_data, self.peaks.best_fit, self.background, ax) # plot the fitted data
        try:
            ax = plotting.plot_uncertainty_curve(self.data[self.x_label], self.peaks.eval_uncertainty(sigma=3),
                                         self.peaks.best_fit, ax) #plot a band of uncertainty
        except TypeError:
            logging.warning('There are not uncertainties available for some reason - '
                         'try lowering the tightness of automatic bounds')
        ax = plotting.plot_data(self.x_data, self.y_data, ax)  # plot the raw data

        ax.minorticks_on()
        ax.grid(which='minor', alpha=0.2)
        ax.grid(which='major', alpha=0.5)

        self.plot = plt

    def fit_report(self):
        self.fit_report = {mod_no:{} for mod_no in range(len(self.peak_types))}

        ## total fit data
        self.fit_report['chi_sq'] = self.chi_sq
        self.fit_report['free_params'] = self.free_params
        self.fit_report['p_value'] = 'not implemented'

        ## individual parameter data
        param_info = {"center":"centers", "amplitude":"amps", "sigma":"widths", "fwhm":False, "height":False}
        for parameter, param_obj in self.peaks.params.items():
            model_no = int(re.findall(r'\d+', parameter)[0])
            param_type = param_info[difflib.get_close_matches(parameter, param_info.keys())[0]]

            if param_type:
                value =param_obj.value
                err = param_obj.stderr
                type = self.peak_types[model_no]
                bounds = self.bounds[param_type][model_no]

                fit_info = {"value":value,
                            "stderr":err,
                            "peak_type":type,
                            "bounds":bounds}

                self.fit_report[model_no][param_type] = fit_info


def main(arguments):
    thunder = Thunder(copy.deepcopy(arguments)) # load object
    return thunder